/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkAdjustBoundary.cxx

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkAdjustBoundary.h"

#include "vtkCellArray.h"
#include "vtkCellData.h"
#include "vtkCellLocator.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkMath.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkStreamingDemandDrivenPipeline.h"
#include "vtkTriangleFilter.h"

vtkStandardNewMacro(vtkAdjustBoundary);

// The following code defines a helper class for performing mesh smoothing
// across the surface of another mesh.
typedef struct _vtkSmoothPoint {
  vtkIdType cellId; // cell
  int subId; // cell sub id
  double p[3]; // parametric coords in cell
} vtkSmoothPoint;

class vtkSmoothPoints { //;prevent man page generation
public:
  vtkSmoothPoints();
  ~vtkSmoothPoints()
    {
    delete [] this->Array;
    };
  vtkIdType GetNumberOfPoints() {return this->MaxId + 1;};
  vtkSmoothPoint *GetSmoothPoint(vtkIdType i) {return this->Array + i;};
  vtkSmoothPoint *InsertSmoothPoint(vtkIdType ptId)
    {
    if ( ptId >= this->Size )
      {
      this->Resize(ptId+1);
      }
    if ( ptId > this->MaxId )
      {
      this->MaxId = ptId;
      }
    return this->Array + ptId;
    }
  vtkSmoothPoint *Resize(vtkIdType sz); //reallocates data
  void Reset() {this->MaxId = -1;};

  vtkSmoothPoint *Array;  // pointer to data
  vtkIdType MaxId;              // maximum index inserted thus far
  vtkIdType Size;               // allocated size of data
  vtkIdType Extend;             // grow array by this amount
};

vtkSmoothPoints::vtkSmoothPoints()
{
  this->MaxId = -1;
  this->Array = new vtkSmoothPoint[1000];
  this->Size = 1000;
  this->Extend = 5000;
}

vtkSmoothPoint *vtkSmoothPoints::Resize(vtkIdType sz)
{
  vtkSmoothPoint *newArray;
  vtkIdType newSize;

  if (sz >= this->Size)
    {
    newSize = this->Size +
      this->Extend*(((sz-this->Size)/this->Extend)+1);
    }
  else
    {
    newSize = sz;
    }

  newArray = new vtkSmoothPoint[newSize];

  memcpy(newArray, this->Array,
         (sz < this->Size ? sz : this->Size) * sizeof(vtkSmoothPoint));

  this->Size = newSize;
  delete [] this->Array;
  this->Array = newArray;

  return this->Array;
}

// The following code defines methods for the vtkAdjustBoundary class
//

// Construct object with number of iterations 20; relaxation factor .01;
// edge angle 15 degrees;
// convergence criterion is 0.0 of the bounding box diagonal.
vtkAdjustBoundary::vtkAdjustBoundary()
{
  this->Convergence = 0.0; //goes to number of specified iterations
  this->NumberOfIterations = 20;

  this->RelaxationFactor = 0.01;

  this->EdgeAngle = 15.0;

  this->SetNumberOfInputPorts(2);
}

//------------------------------------------------------------------------------
void vtkAdjustBoundary::SetSourceData(vtkPolyData *source)
{
  this->SetInputData(1, source);
}

//------------------------------------------------------------------------------
vtkPolyData *vtkAdjustBoundary::GetSource()
{
  if (this->GetNumberOfInputConnections(1) < 1)
    {
    return NULL;
    }
  return vtkPolyData::SafeDownCast(
    this->GetExecutive()->GetInputData(1, 0));
}

//------------------------------------------------------------------------------

#define VTK_SIMPLE_VERTEX 0
#define VTK_FIXED_VERTEX 1
#define VTK_FEATURE_EDGE_VERTEX 2
#define VTK_BOUNDARY_EDGE_VERTEX 3

//------------------------------------------------------------------------------
// Special structure for marking vertices
typedef struct _vtkMeshVertex
{
  char type;
  vtkIdList *edges; // connected edges (list of connected point ids)
} vtkMeshVertex, *vtkMeshVertexPtr;

//------------------------------------------------------------------------------
int vtkAdjustBoundary::RequestData(
  vtkInformation *vtkNotUsed(request),
  vtkInformationVector **inputVector,
  vtkInformationVector *outputVector)
{
  // get the info objects
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *sourceInfo = inputVector[1]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  // get the input and output
  vtkPolyData *input = vtkPolyData::SafeDownCast(
    inInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkPolyData *source = 0;
  source = vtkPolyData::SafeDownCast(
    sourceInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkPolyData *output = vtkPolyData::SafeDownCast(
    outInfo->Get(vtkDataObject::DATA_OBJECT()));

  vtkIdType numPts, numCells, i, j;
  vtkIdType npts;
  vtkPolyData *inMesh, *Mesh;
  vtkTriangleFilter *toTris=NULL;
  double x[3], y[3], deltaX[3], xNew[3];
  double x1[3], x2[3], x3[3], l1[3], l2[3];
  double CosEdgeAngle; // Cosine of angle between adjacent edges
  vtkIdType numSimple=0, numBEdges=0, numFixed=0, numFEdges=0;

  // Check input
  //
  numPts = input->GetNumberOfPoints();
  numCells = input->GetNumberOfCells();
  if (numPts < 1 || numCells < 1)
    {
    vtkErrorMacro(<<"No data to adjust!");
    return 1;
    }

  CosEdgeAngle = cos(vtkMath::RadiansFromDegrees(this->EdgeAngle));

  vtkDebugMacro(<<"Adjusting " << numPts << " vertices, " << numCells
               << " cells with:\n"
               << "\tConvergence= " << this->Convergence << "\n"
               << "\tIterations= " << this->NumberOfIterations << "\n"
               << "\tRelaxation Factor= " << this->RelaxationFactor << "\n"
               << "\tEdge Angle= " << this->EdgeAngle << "\n");

  if (this->NumberOfIterations <= 0 || this->RelaxationFactor == 0.0)
    { //don't do anything! pass data through
    output->CopyStructure(input);
    output->GetPointData()->PassData(input->GetPointData());
    output->GetCellData()->PassData(input->GetCellData());
    return 1;
    }

  // Perform topological analysis. What we're gonna do is built a connectivity
  // array of connected vertices. The outcome will be one of three
  // classifications for a verte: VTK_SIMPLE_VERTEX, VTK_FIXED_VERTEX. or
  // VTK_EDGE_VERTEX. Simple vertices are smoothed using all connected
  // vertices. FIXED vertices are never smoothed. Edge vertices are smoothed
  // using a subset of the attached vertices.
  //
  vtkDebugMacro(<<"Analyzing topology...");
  vtkMeshVertex *Verts = new vtkMeshVertex[numPts];
  for (i = 0; i < numPts; ++i)
    {
    Verts[i].type = VTK_SIMPLE_VERTEX; // can smooth
    Verts[i].edges = NULL;
    }

  vtkPoints *inPts = input->GetPoints();
  double conv = this->Convergence * input->GetLength();

  // polygons and triangle strips
  vtkCellArray *inPolys = input->GetPolys();
  vtkIdType numPolys = inPolys->GetNumberOfCells();
  vtkCellArray *inStrips = input->GetStrips();
  vtkIdType numStrips = inStrips->GetNumberOfCells();

  if ( numPolys > 0 || numStrips > 0 )
    { // build cell structure
    vtkCellArray  *polys;
    vtkIdType cellId;

    vtkIdList *neighbors = vtkIdList::New();
    neighbors->Allocate(VTK_CELL_SIZE);

    inMesh = vtkPolyData::New();
    inMesh->SetPoints(inPts);
    inMesh->SetPolys(inPolys);
    Mesh = inMesh;

    if ( (numStrips = inStrips->GetNumberOfCells()) > 0 )
      { // convert data to triangles
      inMesh->SetStrips(inStrips);
      toTris = vtkTriangleFilter::New();
      toTris->SetInputData(inMesh);
      toTris->Update();
      Mesh = toTris->GetOutput();
      }

    Mesh->BuildLinks(); // to do neighborhood searching
    polys = Mesh->GetPolys();
    this->UpdateProgress(0.375);

    npts = 0;
    vtkIdType *pts = 0;

    for (cellId = 0, polys->InitTraversal(); polys->GetNextCell(npts,pts);
    ++cellId)
      {
      vtkIdType p1, p2;
      for (i = 0; i < npts; ++i)
        {
        p1 = pts[i];
        p2 = pts[(i+1)%npts];

        if (Verts[p1].edges == NULL)
          {
          Verts[p1].edges = vtkIdList::New();
          Verts[p1].edges->Allocate(16,6);
          }
        if (Verts[p2].edges == NULL)
          {
          Verts[p2].edges = vtkIdList::New();
          Verts[p2].edges->Allocate(16,6);
          }

        Mesh->GetCellEdgeNeighbors(cellId,p1,p2,neighbors);
        int numNei = neighbors->GetNumberOfIds();

        int edge = VTK_SIMPLE_VERTEX;
        if (numNei == 0)
          {
          edge = VTK_BOUNDARY_EDGE_VERTEX;
          }
        else if (numNei >= 2)
          {
          // check to make sure that this edge hasn't been marked already
          for (j = 0; j < numNei; ++j)
            {
            if (neighbors->GetId(j) < cellId)
              {
              break;
              }
            }
          if (j >= numNei)
            {
            edge = VTK_FEATURE_EDGE_VERTEX;
            }
          }
        else // a visited edge; skip rest of analysis
          {
          continue;
          }

        if (edge && Verts[p1].type == VTK_SIMPLE_VERTEX)
          {
          Verts[p1].edges->Reset();
          Verts[p1].edges->InsertNextId(p2);
          Verts[p1].type = edge;
          }
        else if ( (edge && Verts[p1].type == VTK_BOUNDARY_EDGE_VERTEX) ||
        (edge && Verts[p1].type == VTK_FEATURE_EDGE_VERTEX) ||
        (!edge && Verts[p1].type == VTK_SIMPLE_VERTEX))
          {
          Verts[p1].edges->InsertNextId(p2);
          if (Verts[p1].type && edge == VTK_BOUNDARY_EDGE_VERTEX)
            {
            Verts[p1].type = VTK_BOUNDARY_EDGE_VERTEX;
            }
          }

        if (edge && Verts[p2].type == VTK_SIMPLE_VERTEX)
          {
          Verts[p2].edges->Reset();
          Verts[p2].edges->InsertNextId(p1);
          Verts[p2].type = edge;
          }
        else if ((edge && Verts[p2].type == VTK_BOUNDARY_EDGE_VERTEX) ||
        (edge && Verts[p2].type == VTK_FEATURE_EDGE_VERTEX) ||
        (!edge && Verts[p2].type == VTK_SIMPLE_VERTEX))
          {
          Verts[p2].edges->InsertNextId(p1);
          if ( Verts[p2].type && edge == VTK_BOUNDARY_EDGE_VERTEX)
            {
            Verts[p2].type = VTK_BOUNDARY_EDGE_VERTEX;
            }
          }
        }
      }

    inMesh->Delete();
    if (toTris)
      {
      toTris->Delete();
      }

    neighbors->Delete();
    }// if strips or polys

  this->UpdateProgress(0.50);

  // post-process edge vertices to make sure we can smooth them
  for (i = 0; i < numPts; ++i)
    {
    if (Verts[i].type == VTK_SIMPLE_VERTEX)
      {
      ++numSimple;
      }
    else if (Verts[i].type == VTK_FIXED_VERTEX)
      {
      ++numFixed;
      }
    else if (Verts[i].type == VTK_FEATURE_EDGE_VERTEX ||
    Verts[i].type == VTK_BOUNDARY_EDGE_VERTEX)
      { // see how many edges; if two, what the angle is

      if ((npts = Verts[i].edges->GetNumberOfIds()) != 2)
        {
        Verts[i].type = VTK_FIXED_VERTEX;
        ++numFixed;
        }
      else // check angle between edges
        {
        inPts->GetPoint(Verts[i].edges->GetId(0),x1);
        inPts->GetPoint(i,x2);
        inPts->GetPoint(Verts[i].edges->GetId(1),x3);

        for (int k = 0; k < 3; ++k)
          {
          l1[k] = x2[k] - x1[k];
          l2[k] = x3[k] - x2[k];
          }
        if (vtkMath::Normalize(l1) >= 0.0 &&
            vtkMath::Normalize(l2) >= 0.0 &&
            vtkMath::Dot(l1,l2) < CosEdgeAngle)
          {
          ++numFixed;
          Verts[i].type = VTK_FIXED_VERTEX;
          }
        else
          {
          if (Verts[i].type == VTK_FEATURE_EDGE_VERTEX)
            {
            ++numFEdges;
            }
          else
            {
            ++numBEdges;
            }
          }
        }
      }
    }

  vtkDebugMacro(<<"Found\t" << numSimple << " simple vertices\n\t"
                << numFEdges << " feature edge vertices\n\t"
                << numBEdges << " boundary edge vertices\n\t"
                << numFixed << " fixed vertices\n\t");

  vtkDebugMacro(<<"Beginning smoothing and adjusting iterations...");

  vtkPoints *newPts = vtkPoints::New();
  newPts->SetNumberOfPoints(numPts);

  // we do adjust (that is, points are constarined to the
  // surface of the mesh object).

  double closestPt[3], dist2, *w = NULL;
  this->SmoothPoints = new vtkSmoothPoints;
  vtkSmoothPoint *sPtr;
  vtkCellLocator *cellLocator = vtkCellLocator::New();
  w = new double[source->GetMaxCellSize()];

  cellLocator->SetDataSet(source);
  cellLocator->BuildLocator();

  for (i = 0; i < numPts; ++i)
    {
    if (Verts[i].type != VTK_BOUNDARY_EDGE_VERTEX)
      {
      newPts->SetPoint(i,inPts->GetPoint(i));
      }
    else
      {
      sPtr = this->SmoothPoints->InsertSmoothPoint(i);
      cellLocator->FindClosestPoint(inPts->GetPoint(i), closestPt,
                                  sPtr->cellId, sPtr->subId, dist2);
      newPts->SetPoint(i, closestPt);
      }
    }

  double factor = this->RelaxationFactor;
  double maxDist = VTK_DOUBLE_MAX;
  int iterationNumber;
  for (iterationNumber = 0;
  maxDist > conv && iterationNumber < this->NumberOfIterations;
  ++iterationNumber)
    {

    if (iterationNumber && !(iterationNumber % 5))
      {
      this->UpdateProgress(0.5 + 0.5 * iterationNumber/this->NumberOfIterations);
      if (this->GetAbortExecute())
        {
        break;
        }
      }

    maxDist = 0.0;
    for (i = 0; i < numPts; ++i)
      {
      if (Verts[i].type == VTK_BOUNDARY_EDGE_VERTEX && Verts[i].edges != NULL &&
      (npts = Verts[i].edges->GetNumberOfIds()) > 0)
        {
        newPts->GetPoint(i, x); //use current points
        deltaX[0] = deltaX[1] = deltaX[2] = 0.0;
        for (j = 0; j < npts; ++j)
          {
          newPts->GetPoint(Verts[i].edges->GetId(j), y);
          for (int k = 0; k < 3; ++k)
            {
            deltaX[k] += (y[k] - x[k]) / npts;
            }
          } // for all connected points

          for (int k = 0; k < 3; ++k)
            {
            xNew[k] = x[k] + factor * deltaX[k];
            }

          // Constrain point to source
          vtkSmoothPoint *sPtr = this->SmoothPoints->GetSmoothPoint(i);
          vtkCell *cell = NULL;

          if (sPtr->cellId >= 0) // in cell
            {
            cell = source->GetCell(sPtr->cellId);
            }

          if (!cell || cell->EvaluatePosition(xNew, closestPt,
          sPtr->subId, sPtr->p, dist2, w) == 0)
            { // not in cell anymore
            cellLocator->FindClosestPoint(xNew, closestPt, sPtr->cellId,
                                          sPtr->subId, dist2);
            }
          for (int k = 0; k < 3; ++k)
            {
            xNew[k] = closestPt[k];
            }

        newPts->SetPoint(i,xNew);
        double dist = vtkMath::Norm(deltaX);
        if (dist > maxDist)
          {
          maxDist = dist;
          }
        }
      }
    }

  vtkDebugMacro(<<"Performed " << iterationNumber << " smoothing passes");
  cellLocator->Delete();
  delete this->SmoothPoints;
  delete [] w;

  // Update output. Only point coordinates have changed.
  //
  output->GetPointData()->PassData(input->GetPointData());
  output->GetCellData()->PassData(input->GetCellData());

  output->SetPoints(newPts);
  newPts->Delete();

  output->SetVerts(input->GetVerts());
  output->SetLines(input->GetLines());
  output->SetPolys(input->GetPolys());
  output->SetStrips(input->GetStrips());

  // free up connectivity storage
  delete [] Verts;

  return 1;
}

//------------------------------------------------------------------------------
int vtkAdjustBoundary::FillInputPortInformation(int port,
                                                vtkInformation *info)
{
  if (!this->Superclass::FillInputPortInformation(port, info))
    {
    return 0;
    }

  if (port == 1)
    {
    info->Set(vtkAlgorithm::INPUT_IS_OPTIONAL(), 1);
    }
  return 1;
}

//------------------------------------------------------------------------------
void vtkAdjustBoundary::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);

  os << indent << "Source: " << static_cast<void*>(this->GetSource()) << "\n";
}