/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkAdjustBoundary.h

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkAdjustBoundary - adjust point positions lie on the boundary.
// .SECTION Description
// vtkAdjustBoundary is a filter that adjusts point coordinates lie on the
// boundary defined by the Source ivar and carries out constrained smoothing on it.

#ifndef vtkAdjustBoundary_h
#define vtkAdjustBoundary_h

#include "vtkPolyDataAlgorithm.h"

class vtkSmoothPoints;

class vtkAdjustBoundary : public vtkPolyDataAlgorithm
{
public:
  vtkTypeMacro(vtkAdjustBoundary,vtkPolyDataAlgorithm);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Construct object.
  static vtkAdjustBoundary *New();

  // Description:
  // Specify a convergence criterion for the iteration
  // process. Smaller number results in more smoothing iterations.
  vtkSetClampMacro(Convergence,double, 0.0,1.0);
  vtkGetMacro(Convergence,double);

  // Description:
  // Specify the number of iterations for Laplacian smoothing,
  vtkSetClampMacro(NumberOfIterations,int,0,VTK_INT_MAX);
  vtkGetMacro(NumberOfIterations,int);

  // Description:
  // Specify the relaxation factor for Laplacian smoothing. As in all
  // iterative methods, the stability of the process is sensitive to
  // this parameter. In general, small relaxation factors and larger
  // numbers of iterations are more stable thatn larger relaxation
  // factors and smaller numbers of iterations.
  vtkSetMacro(RelaxationFactor,double);
  vtkGetMacro(RelaxationFactor,double);

  // Description:
  // Specify the edge angle to control smoothing along edges (either interior
  // or boundary).
  vtkSetClampMacro(EdgeAngle,double,0.0,180.0);
  vtkGetMacro(EdgeAngle,double);

  // Description:
  // Secify the source object which is used to constrain boundary. The
  // source defines a polyline that the input's boundary is constrained
  // lie upon.
  void SetSourceData(vtkPolyData *source);
  vtkPolyData *GetSource();

protected:
  vtkAdjustBoundary();
  ~vtkAdjustBoundary() {}

  virtual int RequestData(vtkInformation *, vtkInformationVector **, vtkInformationVector *);
  virtual int FillInputPortInformation(int port, vtkInformation *info);

  double Convergence;
  int NumberOfIterations;
  double RelaxationFactor;
  double EdgeAngle;

  vtkSmoothPoints *SmoothPoints;
private:
  vtkAdjustBoundary(const vtkAdjustBoundary&);  // Not implemented.
  void operator=(const vtkAdjustBoundary&);  // Not implemented.
};

#endif
